import boto3
import click
from pool_map import multi_process
import logging
from termcolor import colored
from botocore.exceptions import ClientError

public_perms = "http://acs.amazonaws.com/groups/global/AllUsers"


def get_bucket_permission(bucket):
    s3 = boto3.resource('s3')
    bkt = s3.Bucket(bucket)
    bucket_acl = bkt.Acl().grants
    owner = bkt.Acl().owner['ID']
    perms = {'Public': [], 'Canonical': [], 'Owner': []}
    for p in bucket_acl:
        if p['Grantee']['Type'] == "Group":
            if p['Grantee']['URI'] == public_perms:
                perms['Public'].append(p['Permission'])
        if p['Grantee']['Type'] == "CanonicalUser":
            o = p['Grantee']['ID']
            if owner != o:
                data = {
                    'CanonicalID': o,
                    'Permission': p['Permission']
                }
                perms['Canonical'].append(data)
            if owner == o:
                perms['Owner'].append( p['Permission'] )
    print('Bucket: %s' % bucket)
    print("        Owner ID: %s" % owner)
    if perms['Public']:
        print(colored('        Public Access: %s', 'red')) % perms['Public']
    if perms['Canonical']:
        for i in perms['Canonical']:
            print(colored("        AWS Account: %s %s", 'yellow') % (i['CanonicalID'], i['Permission']))
    print(colored('        Bucket Owner: %s', 'green')) % ', '.join(perms['Owner'])
    print('\n')


def get_permission(x):
    s3 = boto3.resource( 's3' )
    bucket, key, owner = x[0], x[1], x[2]
    obj = s3.Object(bucket, key)
    failed = False
    try:
        obj_acl = obj.Acl().grants
    except ClientError as e:
        if e.response['Error']['Message'] == "Access Denied":
            click.echo('Failed to retrieve ACLs for %s with error Access Denied' %key)
            failed = True
            pass
    perms = {'Public': [], 'Canonical': [], 'Owner': []}
    if not failed:
        for p in obj_acl:
            if p['Grantee']['Type'] == "Group":
                if p['Grantee']['URI'] == public_perms:
                    perms['Public'].append( p['Permission'] )
            if p['Grantee']['Type'] == "CanonicalUser":
                o = p['Grantee']['ID']
                if owner != o:
                    perms['Canonical'].append(p['Permission'])
                if owner == o:
                    perms['Owner'].append( p['Permission'] )

        if perms['Public']:
            # out = {'Key': key, 'Permissions': perms}
            out = {'Key': key, "Public": perms['Public'], "Canonical": perms['Canonical'], "Owner": perms['Owner']}
            print("{} | {} | {} | {}".format(", ".join(out['Public']).ljust(25), ", ".join(out['Canonical']).ljust(25), ", ".join(out['Owner']).ljust(25), out['Key'].ljust(50)))
        else:
            logging.warning('Key is Safe: %s' % key)


def scan_key_perms(scanperms, bucket, prefix, workers):
    s3 = boto3.resource( 's3' )
    bkt = s3.Bucket( bucket )
    owner = bkt.Acl().owner['ID']
    click.echo( '>> Scanning bucket ACL' )
    click.echo(30*'=')
    click.echo('>> Scanning objects with PUBLIC ACL')
    print(150 * "-")
    print("Public {} | Other AWS Accounts {} | Owner {}| Key {}").format("".ljust(18), "".ljust(6), "".ljust(20), "".ljust(60))
    print(150*"-")
    objects = []
    iterator = bkt.objects.filter(Prefix=prefix)
    processed = False
    for k in iterator:
        processed = False
        if len(objects) < 1000:
            objects.append( [bucket, k.key, owner] )
        else:
            multi_process(get_permission, objects, workers)
            del objects[:]
            processed = True
    if not processed:
        multi_process(get_permission, objects, workers)



import boto3
import click
import logging
import time, sys


def clean_bucket(bucket, prefix):
    s3 = boto3.resource( 's3' )
    click.echo('I will start the cleaning %s/%s in 10 seconds, you still have the chance to stop' % (bucket, prefix))
    for remaining in range( 10, 0, -1 ):
        sys.stdout.write( "\r" )
        sys.stdout.write( "{:2d} seconds remaining.".format( remaining ) )
        sys.stdout.flush()
        time.sleep( 1 )
    click.echo('\nStart cleaning...')
    bkt = s3.Bucket(bucket)
    iterator = bkt.object_versions.filter( Prefix=prefix )
    objects = []
    for obj in iterator:
        objects.append( {'Key': obj.key, 'VersionId': obj.id} )
        logging.warning('Sending to deletion: %s %s' % (obj.key, obj.id))
        if len(objects) == 1000:
            response = bkt.delete_objects(Delete={'Objects': objects})
            objects = []
    if len(objects) > 0:
        click.echo('You have only %s keys, deleting'%len(objects))
        response = bkt.delete_objects( Delete={'Objects': objects} )
    else:
        print('it seem that you got no keys')




import threading
import time
import logging
from functools import wraps


def rate_limited(max_per_second):
    """
    Decorator that make functions not be called faster than
    """
    lock = threading.Lock()
    min_interval = 1.0 / float( max_per_second )

    def decorate(func):
        last_time_called = [0.0]

        @wraps( func )
        def rate_limited_function(*args, **kwargs):
            lock.acquire()
            elapsed = time.clock() - last_time_called[0]
            left_to_wait = min_interval - elapsed

            if left_to_wait > 0:
                logging.warning('throttled')
                time.sleep( left_to_wait )

            lock.release()

            ret = func( *args, **kwargs )
            last_time_called[0] = time.clock()
            return ret

        return rate_limited_function

    return decorate



from multiprocessing import Pool
import logging


def multi_process(func, data, workers):
    logging.warning('Consuming list with %s workers' % workers)
    p = Pool(workers)
    try:
        # the timeout(.get(9999999) is a workaround for the KeyboardInterrupt. without that it just does not work.
        # Seem to be a bug on multiprocessing. Will investigate it later
        p.map_async(func, data).get(9999999)
        p.close()
    except (KeyboardInterrupt, SystemExit):
        print "Caught KeyboardInterrupt, terminating workers"
    except Exception as e:
        print(e)




import click
import boto3
from termcolor import colored
from s3_restore import *
from s3_empty_bucket import *
from s3_scan_key_perms import *
from s3_tail import *
from __init__ import *

@click.group()
@click.version_option()
def cli():
    """This is your s3 friend(bro), that will help you with bucket iterations.
    Try:\n
    \b
    # s3bro restore --help
    # s3bro purge --help
    # s3tk --help
    
    For more help or detailed information please check:
    https://github.com/rsavordelli/s3bro
    https://pypi.org/project/s3bro/
    
    """
    global s3cli
    s3cli = boto3.client('s3')
    pass


def abort_if_false(ctx, param, value):
    if not value:
        ctx.abort()


@cli.command()
@click.argument('restore', nargs=-1)
@click.option('--bucket', '-b', type=str, help='bucket name', required=True)
@click.option('--prefix','-p', type=str, help=' prefix', default='')
@click.option('--days', '-d', type=int, help='Days to keep the restore', required=True)
@click.option('--type', '-t', type=click.Choice(['Standard', 'Expedited', 'Bulk']), help='restore type (Tier)', required=True)
@click.option('--versions/--no-versions','-v', default=False, help='[--no-versions is DEFAULT] - this option will make the restore to include all versions excluding delete markers')
@click.option('--update-restore-date/--do-not-update-restore-date', '-urd', default=False, help='If passed, it will change the restore date for already restored key')
@click.option('--include', '-in', type=str, multiple=True, help='Only restore keys that matches with a given string, you can add multiples times by passing --include multiple times')
@click.option('--exclude', '-ex', type=str, multiple=True, help='Do not restore if key name matches with a given pattern,'
                                  'you can add multiple patterns by inputting')
@click.option('--workers', type=int, help='How many helpers to include in task, default is 10', default=10)
@click.option('--log-level', type=click.Choice(['INFO', 'ERROR', 'DEBUG', 'WARNING']), help='logging type', default='ERROR')
def restore(restore, bucket, prefix, days, type, versions, update_restore_date, workers, include, exclude, log_level):
    """
    restore glacier objects from s3
    """
    if type == "Expedited":
        print(colored('Note: ', 'yellow') + "Expedited requests will likely be throttled. If you want to avoid this please check: ")
        click.echo('https://docs.aws.amazon.com/AmazonS3/latest/dev/restoring-objects.html#restoring-objects-expedited-capacity')
        click.echo(30*'=')
    loglevel(log_level)
    collect_keys(restore, bucket, prefix, days, type, versions, update_restore_date, workers, include, exclude)


@cli.command()
@click.argument('purge', nargs=-1)
@click.option('--bucket','-b', type=str, help='Bucket name', required=True)
@click.option('--prefix', '-p', type=str, default='', help='prefix name - optional')
@click.option('--yes', is_flag=True, callback=abort_if_false,
              expose_value=False,
              prompt='Are you sure you want to continue with this deletion?', help="first confirmation")
@click.option('--yes-really', is_flag=True, callback=abort_if_false,
              expose_value=False,
              prompt="Are you really sure you want to continue the deletion?", help="second confirmation")
@click.option('--log-level', type=click.Choice(['INFO', 'ERROR', 'DEBUG', 'WARNING']), help='logging type', default='ERROR')
def purge(purge, bucket, prefix, log_level):
    """ 
    delete all the bucket content
    """
    loglevel(log_level)
    clean_bucket(bucket, prefix)


@cli.command('scan-objects')
@click.argument('scan-objects', nargs=-1)
@click.option('--bucket','-b', type=str, help='Bucket name', required=True)
@click.option('--prefix', '-p', type=str, default='', help='prefix name - optional')
@click.option('--workers', type=int, help='How many helpers to include in task, default is 10', default=10)
@click.option('--log-level', type=click.Choice(['INFO', 'ERROR', 'DEBUG', 'WARNING']), help='logging type', default='ERROR')
def scan_objects(scan_objects, bucket, prefix, workers, log_level):
    """
    scan object ACLs
    """
    loglevel(log_level)
    scan_key_perms(scan_objects, bucket, prefix, workers)


@cli.command('scan-bucket')
@click.argument('scan-bucket', nargs=-1)
@click.option('--bucket','-b', type=str, help='Bucket name')
@click.option('--all', '-A', is_flag=True, help="Scan permissions for all your buckets (don't combine -b with -A)")
@click.option('--log-level', type=click.Choice(['INFO', 'ERROR', 'DEBUG', 'WARNING']), help='logging type', default='ERROR')
def scan_bucket(scan_bucket, bucket, all, log_level):
    """
    scan bucket ACLs
    """
    if bucket is None and all is False:
        print('Either choose --bucket or --all option')
        quit()
    if bucket is not None and all is True:
        print('pick one dude, you want to scan one bucket (-b) or all (-A)?')
        quit()
    if all:
        bucket_list = [x['Name'] for x in s3cli.list_buckets()['Buckets']]
        for x in bucket_list:
            get_bucket_permission(x)
    elif bucket is not None:
        get_bucket_permission(bucket)


@cli.command()
@click.argument('tail', nargs=-1)
@click.option('--bucket','-b', type=str, help='Bucket name', required=True)
@click.option('--timeout', '-t', type=int, help='How much time (in minutes) to run, it will destroy '
                                                'the resources created after this time', required=True)
def tail(tail, bucket, timeout):
    """ 
    s3 logs in "real-time" through S3 Events (for puts and deletes only)
    """
    tail_init(bucket, timeout)



import logging


def loglevel(level):
    if "ERROR" == level:
        logging.basicConfig(level=logging.ERROR)
    if 'DEBUG' == level:
        logging.basicConfig(level=logging.DEBUG)
    if 'INFO' == level:
        logging.basicConfig(level=logging.INFO)
    if 'WARNING' == level:
        logging.basicConfig(level=logging.WARNING)



from setuptools import setup, find_packages

setup(
    name='s3bro',
    version='1.7',
    package_dir={'s3bro': 's3bro'},
    packages=['s3bro'],
    author='Rubem de Lima Savordelli',
    author_email='rsavordelli@gmail.com',
    url='https://github.com/rsavordelli/s3bro',
    python_requires='>=2.6, <3',
    long_description=open( 'README.rst' ).read(),
    install_requires=[
        'click',
        'boto3',
        'termcolor'
    ],
    entry_points='''
        [console_scripts]
        s3bro=s3bro.cli:cli
    ''',
)



import boto3
import click
import random
import string
import json
import time
import os
from termcolor import colored


s3 = boto3.client('s3')


def create_queue(bucket, bucket_location):
    click.echo('[ Preparing services ]')
    print(50*'=')
    click.echo('Creating a SQS in %s ...' % bucket_location)
    sqs = boto3.client('sqs', region_name=bucket_location)
    char_set = string.ascii_uppercase + string.digits
    response = sqs.create_queue(
        QueueName=bucket + "_" + bucket_location + "_" + ''.join(random.sample(char_set*6, 6)),
        Attributes={
            'MessageRetentionPeriod': '60'
    })
    url = response['QueueUrl']
    data = response['QueueUrl'].split('/')
    account = data[3]
    q_name = data[4]
    arn = "arn:aws:sqs:%s:%s:%s" % (bucket_location, account, q_name)
    data = {'arn': arn, 'url': url, 'account_id': account}
    click.echo("Queue: %s" % data['arn'])
    return data


def add_queue_permission(q_url, q_arn, bucket_location):
    sqs = boto3.client('sqs', region_name=bucket_location)
    policy = {
        "Version": "2012-10-17",
        "Id": q_arn + "/SQSDefaultPolicy",
        "Statement": [{
            "Sid": "Sid1522338529054",
            "Effect": "Allow",
            "Principal": {
                "Service": "s3.amazonaws.com"
            },
            "Action": "SQS:*",
            "Resource": q_arn
        }]
    }

    sqs.set_queue_attributes(QueueUrl=q_url, Attributes={'Policy': json.dumps(policy), })


# def get_s3_notification(bucket):
#     response = s3.get_bucket_notification_configuration(Bucket=bucket)
#     print(response)


def enable_s3_notification(bucket, queue):
    # get_s3_notification(bucket)
    click.echo('Enabling s3 event (s3tail-Event) notification on %s ...' % bucket)
    response = s3.put_bucket_notification_configuration(
        Bucket=bucket,
        NotificationConfiguration={
            'QueueConfigurations': [
                {
                    'Id': 's3tail-Event',
                    'QueueArn': queue,
                    'Events': ['s3:ObjectCreated:*', 's3:ObjectRemoved:*']

                }
            ]
        }
    )
    print(50*'=')
    return response


def delete_resources(q_url, bucket, bucket_location):
    sqs = boto3.client('sqs', region_name=bucket_location)
    print( 30 * '-' )
    click.echo( 'Time is over, deleting resources' )
    click.echo( 'Deleting queue' )
    response = sqs.delete_queue( QueueUrl=q_url )
    if response['ResponseMetadata']['HTTPStatusCode'] != 200:
        click.echo( 'Failed to delete the Queue, please manually delete it to avoid extra-cost' )
    click.echo( 'Disabling Event on S3' )
    response = s3.put_bucket_notification_configuration( Bucket=bucket, NotificationConfiguration={} )
    if response['ResponseMetadata']['HTTPStatusCode'] != 200:
        click.echo( 'Failed to delete s3 notification, please manually disable to avoid extra-cost' )


def parsed_output(msg):
    event_time = msg.get('eventTime')
    event_name = msg.get('eventName')
    user_identity = msg.get('userIdentity').get('principalId')
    ip_address = msg.get('requestParameters').get('sourceIPAddress')
    x_amz_id = msg.get('responseElements').get('x-amz-request-id')
    host_id = msg.get('responseElements').get('x-amz-id-2')
    key = msg.get('s3').get('object').get('key')
    size = msg.get('s3').get('object').get('size')
    v_id = msg.get('s3').get('object').get('versionId')
    return [ip_address, event_time, key, str(size), str(v_id), event_name, user_identity, x_amz_id, host_id]


def s3tail(q_url, min, bucket, bucket_location):
    sqs = boto3.client('sqs', region_name=bucket_location)

    timeout = time.time() + 60 * min
    try:
        while True:
            if time.time() > timeout:
                delete_resources(q_url, bucket, bucket_location)
                break

            response = sqs.receive_message(
                QueueUrl=q_url,
                MaxNumberOfMessages=10,
                VisibilityTimeout=65,
            )
            try:
                raw_msg = response['Messages']
                for m in raw_msg:
                    msg = json.loads( m.get('Body'))
                    if msg.get('Records') is not None:
                        notification = msg.get('Records')[0]
                        print(' '.join(parsed_output(notification)))
            except KeyError as e:
                pass

    except (KeyboardInterrupt, SystemExit):
        print("Caught KeyboardInterrupt, terminating workers")
        delete_resources(q_url, bucket, bucket_location)


def resource_confirmation():
    print(colored("s3bro will create an SQS and enable Event notification in your S3 bucket\n"
                  "After the script is done it will delete the resources automatically.\n"
                  "If you cancel or interrupt the script, the resource deletion may fail.\n"
                  "In that case, you have to manually delete the resources, ok?\n"
                  "To prove you've read the message, type 'agreed' in the confirmation below", 'yellow'))
    res = raw_input('confirmation: ')
    if res != 'agreed':
        print('Read the above message entirely')
        resource_confirmation()
    else:
        os.system( 'cls' if os.name == 'nt' else 'clear' )


def get_bucket_location(bucket):
    get_location = s3.get_bucket_location(Bucket=bucket)
    if get_location['LocationConstraint'] is None:
        bucket_location = 'us-east-1'
    else:
        bucket_location = get_location['LocationConstraint']
    return bucket_location


def tail_init(bucket, timeout):
    bucket_location = get_bucket_location(bucket)
    resource_confirmation()
    queue = create_queue(bucket, bucket_location)
    add_queue_permission(queue['url'], queue['arn'], bucket_location)
    enable_s3_notification(bucket, queue['arn'])
    s3tail(queue['url'], timeout, bucket, bucket_location)








import click
import boto3
import logging
from pool_map import multi_process
from ratelimit import *
from botocore.exceptions import ClientError


@rate_limited(50)
def restore_versions(x):
    s3 = boto3.client('s3')
    s3r = boto3.resource('s3')
    bucket, key, v_id, days, type, update_restore_date = x[0], x[1], x[2], x[3], x[4], x[5]
    obj = s3r.ObjectVersion( bucket, key, v_id )
    try:
        data = obj.head()
        try:
            if data['ResponseMetadata']['HTTPHeaders']['x-amz-storage-class'] == 'GLACIER':
                header = data['ResponseMetadata']['HTTPHeaders']
                if 'x-amz-restore' not in header:
                    click.echo('Submitting restoration request: %s for %s days' % (key, days))
                    response = s3.restore_object(
                        Bucket=bucket,
                        Key=key,
                        VersionId=v_id,
                        RestoreRequest={
                            'Days': days,
                            'GlacierJobParameters': {'Tier': type}
                        }
                    )
                elif 'ongoing-request="true"' in header.values():
                    click.echo('Restoration in-progress: %s VersionID: %s' % (key, v_id))
                elif any( 'ongoing-request="false"' in x for x in header.values() ):
                    if update_restore_date:
                        click.echo('[Updating date] - Restoration complete: %s VersionID: %s for %s days from now' % (key, v_id, days) )
                        response = s3.restore_object(
                            Bucket=bucket,
                            Key=key,
                            VersionId=v_id,
                            RestoreRequest={
                                'Days': days,
                                'GlacierJobParameters':
                                    {'Tier': type}
                            }
                        )
                    else:
                        date = ''.join( data['ResponseMetadata']['HTTPHeaders']['x-amz-restore'].split( ',' )[1:] )
                        click.echo('Restoration completed: %s VersionID: %s until %s' % (key, v_id, date))

        except KeyError as e:
            # that's for the keys without x-amz-storage-class header
            logging.warning( '[Key not in glacier] - Key: %s/%s VersionID: %s' % (bucket, key, v_id) )
            pass

    except ClientError as e:
        # print(e.response)
        if e.response['ResponseMetadata']['HTTPStatusCode'] == 503:
            err = {
                'key': key,
                'error': e.response['Error']
            }
            click.echo('ERROR: %s' % err)
            pass
        elif e.response['ResponseMetadata']['HTTPStatusCode'] == 405:
            logging.warning( '[Delete Marker Found] - Ignoring key %s/%s VersionID: %s' % (bucket, key, v_id) )
            pass
        else:
            click.log(e.response)
            pass
    except Exception as e:
        print(e)


@rate_limited(50)
def restore_default(x):
    s3r = boto3.resource( 's3' )
    bucket, key, days, type, update_restore_date = x[0], x[1], x[2], x[3], x[4]
    obj = s3r.Object(bucket, key)
    try:
        if obj.storage_class == 'GLACIER':
            if obj.restore is None:
                click.echo('Submitting restoration request: %s' % obj.key)
                obj.restore_object( RestoreRequest={'Days': days, 'GlacierJobParameters': {'Tier': type}} )
            # Print out objects whose restoration is on-going
            elif 'ongoing-request="true"' in obj.restore:
                click.echo('Restoration in-progress: %s' % obj.key)
            # Print out objects whose restoration is complete
            elif 'ongoing-request="false"' in obj.restore:
                date = obj.restore.split('expiry-date=')[1]
                if update_restore_date:
                    click.echo('[Updating date] - Restoration complete: %s for %s days from now' % (obj.key, days))
                    obj.restore_object( RestoreRequest={'Days': days, 'GlacierJobParameters': {'Tier': type}} )
                else:
                    click.echo('Restoration completed: %s until %s' % (obj.key, date))
        else:
            logging.warning('[Key not in glacier] - Key: %s' % obj.key)
    except ClientError as e:
        err = {'key': key, 'error': e.response['Error']}
        if e.response['ResponseMetadata']['HTTPStatusCode'] == 503:
            click.echo('ERROR: %s' % err)
            pass
        else:
            click.echo(err)
    except KeyError as e:
        click.echo(e.response)
    except Exception as e:
        click.echo(e)


def collect_keys(restore, bucket, prefix, days, type, versions, update_restore_date, workers, include, exclude):
    s3r = boto3.resource( 's3' )
    startTime = time.time()
    bkt = s3r.Bucket(bucket)
    objects = []
    keys_proccessed = 0
    click.echo('Initiating %s restore for %s/%s...\nRestoring keys for %s days\n' %(type, bucket, prefix, days) + 30*'=')
    if versions:
        iterator = bkt.object_versions.filter(Prefix=prefix)
        click.echo('All versions: True\n' + 30*'=')
    else:
        click.echo('All versions: False\n' + 30*'=')
        iterator = bkt.objects.filter( Prefix=prefix )
    processed = False
    for obj in iterator:
        keys_proccessed += 1
        processed = False
        if len(objects) < 1000:
            if versions:
                data = [bucket, obj.key, obj.id, days, type, update_restore_date]
                if include:
                    # if obj.key.endswith(include):
                    if any( x in obj.key for x in include ):
                        objects.append(data)
                    else:
                        logging.warning( 'Not included because of filter (--include) to the key %s | Keys to include %s ' % ( obj.key, include) )
                elif exclude:
                    if not any( x in obj.key for x in exclude ):
                        objects.append(data)
                    else:
                        logging.warning('filter applied (exclude) to the key %s | Keys to exclude %s ' % ( obj.key, exclude) )
                else:
                    objects.append( data )
            else:
                data = [bucket, obj.key, days, type, update_restore_date]
                if include:
                    if any( x in obj.key for x in include ):
                        objects.append(data)
                    else:
                        logging.warning( ' Not included because of filter (--include) to the key %s | Keys to include %s ' % (obj.key, include) )
                elif exclude:
                    if not any( x in obj.key for x in exclude ):
                        objects.append(data)
                    else:
                        logging.warning( 'filter applied (exclude) to the key %s | Keys to exclude %s ' % (obj.key, exclude) )
                else:
                    objects.append( data )
        else:
            if versions:
                multi_process(restore_versions, objects, workers)
            else:
                multi_process(restore_default, objects, workers)
            # click.echo('cleaning objects')
            del objects[:]
            processed = True

    if not processed:
        # click.echo('There is only %s keys, running it in a single batch\n'%(len(objects)) + 30*'=')
        if versions:
            multi_process(restore_versions, objects, workers)
        else:
            multi_process(restore_default, objects, workers)
    elapsed = time.time() - startTime
    end = round(elapsed, 2)
    click.echo('Total keys proccessed: %s in %ss' %(keys_proccessed, end))